#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>

#include <event.h>

void cb_read(int fd, short event, void *arg);

class Connection {
public:
    int m_socket;
    struct event m_event;

    Connection(int socket) : m_socket(socket) {}
    virtual ~Connection() {
        event_del(&m_event);
    }
};

void
cb_accept(int fd, short event, void *arg)
{
    struct event *ev = (struct event *)arg;
    event_add(ev, NULL);

    struct sockaddr_in sin;
    socklen_t sinsiz = sizeof(sin);
    int s;

    s = accept(fd, (struct sockaddr *)&sin, &sinsiz);

    if (s == -1) {
        return;
    }

    {
        int flags;
        flags = fcntl(s, F_GETFL, 0);
        fcntl(s, F_SETFL, flags | O_NONBLOCK);
    }

    Connection *newconn = new Connection(s);

    event_set(&newconn->m_event, s, EV_READ, cb_read, newconn);
    event_add(&newconn->m_event, NULL);
}

void
cb_read(int fd, short event, void *arg)
{
    Connection *conn = (Connection *)arg;
    struct event *ev = &conn->m_event;

    char buf[255];
    int len;

    if (event & EV_READ) {
        len = read(fd, buf, sizeof(buf) - 1);

        if (len == -1) {
            printf("read error (len=%d)\n", len);
            return;
        } else if (len == 0) {
            printf("connection closed\n");
            event_del(ev);
            delete conn;
            return;
        }

        buf[len] = '\0';

        write(fd, buf, len);
        event_add(ev, NULL);
    }
}

int
main (register int const argc, register char *const argv[])
{
    struct in_addr listen_addr;
    int one = 1;
    unsigned int portno = 2007;
    int server_sock;
    struct sockaddr_in sin;

    listen_addr.s_addr = htonl (INADDR_ANY);  /* Default.  */

    if ((server_sock = socket (PF_INET, SOCK_STREAM, 0)) == -1) {
        printf("Error creating socket: %s\n", strerror(errno));
        exit(1);
    }

    if (setsockopt(server_sock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof one) == -1) {
        printf("Error setting SO_REUSEADDR for socket: %s\n", strerror (errno));
        exit(1);
    }

    memset (&sin, 0, sizeof sin);
    sin.sin_family = AF_INET;
    sin.sin_addr = listen_addr;
    sin.sin_port = htons (portno);

    if (bind (server_sock, (const struct sockaddr *)&sin, sizeof sin) == -1) {
        printf("Error binding socket: %s\n", strerror (errno));
        exit(1);
    }

    if (listen (server_sock, 20) == -1) {
        printf("Error listening to socket: %s\n", strerror (errno));
        exit(1);
    }

    {
        int flags;
        flags = fcntl(server_sock, F_GETFL, 0);
        fcntl(server_sock, F_SETFL, flags | O_NONBLOCK);
    }

    event_init();
    struct event listen_ev;
    struct event remove_old_user_ev;

    printf("libevent: %s\n", event_get_method());

    event_set(&listen_ev, server_sock, EV_READ, cb_accept, &listen_ev);
    event_add(&listen_ev, NULL);

    event_dispatch();
}
